import numpy as np
from unittest.mock import patch, ANY

from .base_test_class import DartsBaseTestClass
from ..utils import timeseries_generation as tg
from ..metrics import mape
from ..logging import get_logger
from ..dataprocessing.transformers import Scaler
from ..datasets import AirPassengersDataset
from darts.models.forecasting.torch_forecasting_model import DualCovariatesTorchModel
from darts.utils.timeseries_generation import linear_timeseries

logger = get_logger(__name__)

try:
    from ..models import BlockRNNModel, TCNModel, TransformerModel, NBEATSModel, RNNModel
    from darts.utils.likelihood_models import GaussianLikelihood
    import torch
    TORCH_AVAILABLE = True
except ImportError:
    logger.warning('Torch not installed - will be skipping Torch models tests')
    TORCH_AVAILABLE = False


if TORCH_AVAILABLE:
    IN_LEN = 24
    OUT_LEN = 12
    models_cls_kwargs_errs = [
        (BlockRNNModel, {'model': 'RNN', 'hidden_size': 10, 'n_rnn_layers': 1, 'batch_size': 32, 'n_epochs': 10}, 180.),
        (RNNModel, {'model': 'RNN', 'hidden_dim': 10, 'batch_size': 32, 'n_epochs': 10}, 180.),
        (RNNModel, {'training_length': 12, 'n_epochs': 10, 'likelihood': GaussianLikelihood()}, 80),
        (TCNModel, {'n_epochs': 10, 'batch_size': 32}, 240.),
        (TransformerModel, {'d_model': 16, 'nhead': 2, 'num_encoder_layers': 2, 'num_decoder_layers': 2,
                            'dim_feedforward': 16, 'batch_size': 32, 'n_epochs': 10}, 180.),
        (NBEATSModel, {'num_stacks': 4, 'num_blocks': 1, 'num_layers': 2, 'layer_widths': 12, 'n_epochs': 10}, 180.)
    ]

    class GlobalForecastingModelsTestCase(DartsBaseTestClass):
        # forecasting horizon used in runnability tests
        forecasting_horizon = 12

        np.random.seed(42)
        torch.manual_seed(42)

        # real timeseries for functionality tests
        ts_passengers = AirPassengersDataset().load()
        scaler = Scaler()
        ts_passengers = scaler.fit_transform(ts_passengers)
        ts_pass_train, ts_pass_val = ts_passengers[:-36], ts_passengers[-36:]

        # an additional noisy series
        ts_pass_train_1 = ts_pass_train + 0.01 * tg.gaussian_timeseries(length=len(ts_pass_train),
                                                                        freq=ts_pass_train.freq_str,
                                                                        start=ts_pass_train.start_time())

        # an additional time series serving as covariates
        year_series = tg.datetime_attribute_timeseries(ts_passengers, attribute='year')
        month_series = tg.datetime_attribute_timeseries(ts_passengers, attribute='month')
        scaler_dt = Scaler()
        time_covariates = scaler_dt.fit_transform(year_series.stack(month_series))
        time_covariates_train, time_covariates_val = time_covariates[:-36], time_covariates[-36:]

        # an artificial time series that is highly dependent on covariates
        ts_length = 400
        split_ratio = 0.6
        sine_1_ts = tg.sine_timeseries(length=ts_length)
        sine_2_ts = tg.sine_timeseries(length=ts_length, value_frequency=0.05)
        sine_3_ts = tg.sine_timeseries(length=ts_length, value_frequency=0.003, value_amplitude=5)
        linear_ts = tg.linear_timeseries(length=ts_length, start_value=3, end_value=8)

        covariates = sine_3_ts.stack(sine_2_ts).stack(linear_ts)
        covariates_past, _ = covariates.split_after(split_ratio)

        target = sine_1_ts + sine_2_ts + linear_ts + sine_3_ts
        target_past, target_future = target.split_after(split_ratio)

        def test_single_ts(self):
            for model_cls, kwargs, err in models_cls_kwargs_errs:
                model = model_cls(input_chunk_length=IN_LEN, output_chunk_length=OUT_LEN, random_state=0, **kwargs)
                model.fit(self.ts_pass_train)
                pred = model.predict(n=36)
                mape_err = mape(self.ts_pass_val, pred)
                self.assertTrue(mape_err < err, 'Model {} produces errors too high (one time '
                                                'series). Error = {}'.format(model_cls, mape_err))

        def test_multi_ts(self):
            for model_cls, kwargs, err in models_cls_kwargs_errs:
                model = model_cls(input_chunk_length=IN_LEN, output_chunk_length=OUT_LEN, random_state=0, **kwargs)
                model.fit([self.ts_pass_train, self.ts_pass_train_1])
                with self.assertRaises(ValueError):
                    # when model is fit from >1 series, one must provide a series in argument
                    model.predict(n=1)
                pred = model.predict(n=36, series=self.ts_pass_train)
                mape_err = mape(self.ts_pass_val, pred)
                self.assertTrue(mape_err < err, 'Model {} produces errors too high (several time '
                                'series). Error = {}'.format(model_cls, mape_err))

                # check prediction for several time series
                pred_list = model.predict(n=36, series=[self.ts_pass_train, self.ts_pass_train_1])
                self.assertTrue(len(pred_list) == 2, 'Model {} did not return a list of prediction'.format(model_cls))
                for pred in pred_list:
                    mape_err = mape(self.ts_pass_val, pred)
                    self.assertTrue(mape_err < err, 'Model {} produces errors too high (several time series 2). '
                                                    'Error = {}'.format(model_cls, mape_err))

        def test_covariates(self):
            for model_cls, kwargs, err in models_cls_kwargs_errs:

                model = model_cls(input_chunk_length=IN_LEN, output_chunk_length=OUT_LEN, random_state=0, **kwargs)

                # Here we rely on the fact that all non-Dual models currently are Past models
                cov_name = 'future_covariates' if isinstance(model, DualCovariatesTorchModel) else 'past_covariates'
                cov_kwargs = {cov_name: [self.time_covariates_train, self.time_covariates_train]}

                model.fit(series=[self.ts_pass_train, self.ts_pass_train_1], **cov_kwargs)
                with self.assertRaises(ValueError):
                    # when model is fit from >1 series, one must provide a series in argument
                    model.predict(n=1)

                with self.assertRaises(ValueError):
                    # when model is fit using multiple covariates, covariates are required at prediction time
                    model.predict(n=1, series=self.ts_pass_train)

                cov_kwargs_train = {cov_name: self.time_covariates_train}
                cov_kwargs_notrain = {cov_name: self.time_covariates}
                with self.assertRaises(ValueError):
                    # when model is fit using covariates, n cannot be greater than output_chunk_length...
                    model.predict(n=13, series=self.ts_pass_train, **cov_kwargs_train)

                # ... unless future covariates are provided
                pred = model.predict(n=13, series=self.ts_pass_train, **cov_kwargs_notrain)

                pred = model.predict(n=12, series=self.ts_pass_train, **cov_kwargs_notrain)
                mape_err = mape(self.ts_pass_val, pred)
                self.assertTrue(mape_err < err, 'Model {} produces errors too high (several time '
                                                'series with covariates). Error = {}'.format(model_cls, mape_err))

                # when model is fit using 1 training and 1 covariate series, time series args are optional
                if model._is_probabilistic:
                    continue
                model = model_cls(input_chunk_length=IN_LEN, output_chunk_length=OUT_LEN, **kwargs)
                model.fit(series=self.ts_pass_train, **cov_kwargs_train)
                pred1 = model.predict(1)
                pred2 = model.predict(1, series=self.ts_pass_train)
                pred3 = model.predict(1, **cov_kwargs_train)
                pred4 = model.predict(1, **cov_kwargs_train, series=self.ts_pass_train)
                self.assertEqual(pred1, pred2)
                self.assertEqual(pred1, pred3)
                self.assertEqual(pred1, pred4)

        def test_future_covariates(self):
            # models with future covariates should produce better predictions over a long forecasting horizon
            # than a model trained with no covariates
            model = TCNModel(input_chunk_length=50, output_chunk_length=5, n_epochs=20, random_state=0)

            model.fit(series=self.target_past)
            long_pred_no_cov = model.predict(n=160)

            model = TCNModel(input_chunk_length=50, output_chunk_length=5, n_epochs=20, random_state=0)
            model.fit(series=self.target_past, past_covariates=self.covariates_past)
            long_pred_with_cov = model.predict(n=160, past_covariates=self.covariates)
            self.assertTrue(mape(self.target_future, long_pred_no_cov) > mape(self.target_future, long_pred_with_cov),
                            'Models with future covariates should produce better predictions.')

            # block models can predict up to self.output_chunk_length points beyond the last future covariate...
            model.predict(n=165, past_covariates=self.covariates)

            # ... not more
            with self.assertRaises(ValueError):
                model.predict(n=166, series=self.ts_pass_train)

            # recurrent models can only predict data points for time steps where future covariates are available
            model = RNNModel(n_epochs=1)
            model.fit(series=self.target_past, future_covariates=self.covariates_past)
            model.predict(n=160, future_covariates=self.covariates)
            with self.assertRaises(ValueError):
                model.predict(n=161, future_covariates=self.covariates)

        def test_batch_predictions(self):
            # predicting multiple time series at once needs to work for arbitrary batch sizes
            # univariate case
            targets_univar = [self.target_past, self.target_past[:60], self.target_past[:80]]
            self._batch_prediction_test_helper_function(targets_univar)

            # multivariate case
            targets_multivar = [tgt.stack(tgt) for tgt in targets_univar]
            self._batch_prediction_test_helper_function(targets_multivar)

        def _batch_prediction_test_helper_function(self, targets):
            epsilon = 1e-4
            model = TCNModel(input_chunk_length=50, output_chunk_length=10, n_epochs=10, random_state=0)
            model.fit(series=targets[0], past_covariates=self.covariates_past)
            preds_default = model.predict(n=160, series=targets,
                                          past_covariates=[self.covariates] * len(targets), batch_size=None)

            # make batch size large enough to test stacking samples
            for batch_size in range(1, 4 * len(targets)):
                preds = model.predict(n=160, series=targets,
                                      past_covariates=[self.covariates] * len(targets), batch_size=batch_size)

                for i in range(len(targets)):
                    self.assertLess(sum(sum((preds[i] - preds_default[i]).values())), epsilon)

        def test_predict_from_dataset_unsupported_input(self):
            # an exception should be thrown if an unsupported type is passed
            unsupported_type = 'unsupported_type'
            # just need to test this with one model
            model_cls, kwargs, err = models_cls_kwargs_errs[0]
            model = model_cls(input_chunk_length=IN_LEN, output_chunk_length=OUT_LEN, **kwargs)
            model.fit([self.ts_pass_train, self.ts_pass_train_1])

            with self.assertRaises(ValueError):
                model.predict_from_dataset(n=1, input_series_dataset=unsupported_type)

        def test_same_result_with_different_n_jobs(self):
            for model_cls, kwargs, err in models_cls_kwargs_errs:
                model = model_cls(input_chunk_length=IN_LEN, output_chunk_length=OUT_LEN, **kwargs)
                if model._is_probabilistic():
                    continue
                multiple_ts = [self.ts_pass_train] * 10

                model.fit(multiple_ts)

                pred1 = model.predict(n=36, series=multiple_ts, n_jobs=1)
                pred2 = model.predict(n=36, series=multiple_ts, n_jobs=-1)  # assuming > 1 core available in the machine

                self.assertEqual(pred1, pred2, 'Model {} produces different predictions with different number of jobs')

        @patch('darts.models.forecasting.torch_forecasting_model.torch.save')
        @patch('darts.models.forecasting.torch_forecasting_model.TorchForecastingModel._train')
        @patch('darts.models.forecasting.torch_forecasting_model.shutil.rmtree')
        def test_fit_with_constr_epochs(self, rmtree_patch, train_patch, save_patch):
            for model_cls, kwargs, err in models_cls_kwargs_errs:
                model = model_cls(input_chunk_length=IN_LEN, output_chunk_length=OUT_LEN, **kwargs)
                multiple_ts = [self.ts_pass_train] * 10
                model.fit(multiple_ts)

                train_patch.assert_called_with(ANY, ANY, ANY, ANY, kwargs['n_epochs'])

        @patch('darts.models.forecasting.torch_forecasting_model.torch.save')
        @patch('darts.models.forecasting.torch_forecasting_model.TorchForecastingModel._train')
        @patch('darts.models.forecasting.torch_forecasting_model.shutil.rmtree')
        def test_fit_with_fit_epochs(self, rmtree_patch, train_patch, save_patch):
            for model_cls, kwargs, err in models_cls_kwargs_errs:
                model = model_cls(input_chunk_length=IN_LEN, output_chunk_length=OUT_LEN, **kwargs)
                multiple_ts = [self.ts_pass_train] * 10
                epochs = 42

                model.fit(multiple_ts, epochs=epochs)

                train_patch.assert_called_with(ANY, ANY, ANY, ANY, epochs)

                model.total_epochs = epochs
                # continue training
                model.fit(multiple_ts, epochs=epochs)

                train_patch.assert_called_with(ANY, ANY, ANY, ANY, epochs)

        @patch('darts.models.forecasting.torch_forecasting_model.torch.save')
        @patch('darts.models.forecasting.torch_forecasting_model.TorchForecastingModel._train')
        @patch('darts.models.forecasting.torch_forecasting_model.shutil.rmtree')
        def test_fit_from_dataset_with_epochs(self, rmtree_patch, train_patch, save_patch):
            for model_cls, kwargs, err in models_cls_kwargs_errs:
                model = model_cls(input_chunk_length=IN_LEN, output_chunk_length=OUT_LEN, **kwargs)
                multiple_ts = [self.ts_pass_train] * 10
                train_dataset = model._build_train_dataset(multiple_ts, past_covariates=None, future_covariates=None)
                epochs = 42

                model.fit_from_dataset(train_dataset, epochs=epochs)

                train_patch.assert_called_with(ANY, ANY, ANY, ANY, epochs)

                model.total_epochs = epochs
                # continue training
                model.fit_from_dataset(train_dataset, epochs=epochs)

                train_patch.assert_called_with(ANY, ANY, ANY, ANY, epochs)

        def test_sample_smaller_than_batch_size(self):
            """
            Checking that the TorchForecastingModels do not crash even if the number of available samples for training
            is strictly lower than the selected batch_size
            """
            # TS with 50 timestamps. TorchForecastingModels will use the SequentialDataset for producing training
            # samples, which means we will have 50 - 22 - 2 + 1 = 27 samples, which is < 32 (batch_size). The model
            # should still train on those samples and not crash in any way
            ts = linear_timeseries(0, 1, 50)

            model = RNNModel(input_chunk_length=20,
                             output_chunk_length=2,
                             n_epochs=2,
                             batch_size=32)
            model.fit(ts)



"""
TimeSeries Datasets
-------------------
"""

# Base classes for training datasets:
from .training_dataset import (TrainingDataset, PastCovariatesTrainingDataset,
                               FutureCovariatesTrainingDataset, DualCovariatesTrainingDataset,
                               MixedCovariatesTrainingDataset, SplitCovariatesTrainingDataset)

# Base class and implementations for inference datasets:
from .inference_dataset import (InferenceDataset, PastCovariatesInferenceDataset,
                                FutureCovariatesInferenceDataset, DualCovariatesInferenceDataset,
                                MixedCovariatesInferenceDataset, SplitCovariatesInferenceDataset)

# Implementations (sequential)
from .sequential_dataset import (PastCovariatesSequentialDataset, FutureCovariatesSequentialDataset,
                                 DualCovariatesSequentialDataset, MixedCovariatesSequentialDataset,
                                 SplitCovariatesSequentialDataset)

# Implementations (shifted)
from .shifted_dataset import (PastCovariatesShiftedDataset, FutureCovariatesShiftedDataset,
                              DualCovariatesShiftedDataset, MixedCovariatesShiftedDataset,
                              SplitCovariatesShiftedDataset)

# Implementation (horizon-based)
from .horizon_based_dataset import HorizonBasedDataset



import numpy as np
import pandas as pd

from .base_test_class import DartsBaseTestClass
from ..timeseries import TimeSeries
from .test_timeseries import TimeSeriesTestCase


class TimeSeriesMultivariateTestCase(DartsBaseTestClass):

    times1 = pd.date_range('20130101', '20130110')
    times2 = pd.date_range('20130206', '20130215')
    dataframe1 = pd.DataFrame({
        "0": range(10),
        "1": range(5, 15),
        "2": range(10, 20)
    }, index=times1)
    dataframe2 = pd.DataFrame({
        "0": np.arange(1, 11),
        "1": np.arange(1, 11) * 3,
        "2": np.arange(1, 11) * 5
    }, index=times1)
    dataframe3 = pd.DataFrame({
        "0": np.arange(1, 11),
        "1": np.arange(11, 21),
    }, index=times2)
    series1 = TimeSeries.from_dataframe(dataframe1)
    series2 = TimeSeries.from_dataframe(dataframe2)
    series3 = TimeSeries.from_dataframe(dataframe3)

    def test_creation(self):
        series_test = TimeSeries.from_dataframe(self.dataframe1)
        self.assertTrue(np.all(series_test.pd_dataframe().values == (self.dataframe1.values)))

        # Series cannot be lower than three without passing frequency as argument to constructor
        with self.assertRaises(ValueError):
            TimeSeries(self.dataframe1.iloc[:2, :])
        TimeSeries.from_dataframe(self.dataframe1.iloc[:2, :], freq='D')

    def test_eq(self):
        seriesA = TimeSeries.from_dataframe(self.dataframe1)
        self.assertTrue(self.series1 == seriesA)
        self.assertFalse(self.series1 != seriesA)

        # with different dates
        dataframeB = self.dataframe1.copy()
        dataframeB.index = pd.date_range('20130102', '20130111')
        seriesB = TimeSeries.from_dataframe(dataframeB)
        self.assertFalse(self.series1 == seriesB)

        # with one different value
        dataframeC = self.dataframe1.copy()
        dataframeC.iloc[2, 2] = 0
        seriesC = TimeSeries.from_dataframe(dataframeC)
        self.assertFalse(self.series1 == seriesC)

    def test_rescale(self):
        with self.assertRaises(ValueError):
            self.series1.rescale_with_value(1)

        seriesA = self.series2.rescale_with_value(0)
        self.assertTrue(np.all(seriesA.values() == 0).all())

        seriesB = self.series2.rescale_with_value(1)
        self.assertEqual(seriesB, TimeSeries.from_dataframe(
            pd.DataFrame({
                "0": np.arange(1, 11),
                "1": np.arange(1, 11),
                "2": np.arange(1, 11)
            }, index=self.dataframe2.index).astype(float)
        ))

    def test_slice(self):
        TimeSeriesTestCase.helper_test_slice(self, self.series1)

    def test_split(self):
        TimeSeriesTestCase.helper_test_split(self, self.series1)

    def test_drop(self):
        TimeSeriesTestCase.helper_test_drop(self, self.series1)

    def test_intersect(self):
        TimeSeriesTestCase.helper_test_intersect(self, self.series1)

    def test_shift(self):
        TimeSeriesTestCase.helper_test_shift(self, self.series1)

    def test_append(self):
        TimeSeriesTestCase.helper_test_append(self, self.series1)

    def test_strip(self):
        dataframe1 = pd.DataFrame({
            "0": 2 * [np.nan] + list(range(7)) + [np.nan],
            "1": [np.nan] + list(range(7)) + 2 * [np.nan]
        }, index=self.times1)
        series1 = TimeSeries.from_dataframe(dataframe1)

        self.assertTrue((series1.strip().time_index == self.times1[1:-1]).all())

    """
    Testing new multivariate methods.
    """

    def test_stack(self):
        with self.assertRaises(ValueError):
            self.series1.stack(self.series3)
        seriesA = self.series1.stack(self.series2)
        dataframeA = pd.concat([self.dataframe1, self.dataframe2], axis=1)
        dataframeA.columns = ["0", "1", "2", "0_1", "1_1", "2_1"]  # the names to expect after stacking
        self.assertTrue((seriesA.pd_dataframe() == dataframeA).all().all())
        self.assertEqual(seriesA.values().shape, (len(self.dataframe1),
                                                  len(self.dataframe1.columns) + len(self.dataframe2.columns)))

    def test_univariate_component(self):
        with self.assertRaises(IndexError):
            self.series1.univariate_component(-5)
        with self.assertRaises(IndexError):
            self.series1.univariate_component(3)
        seriesA = self.series1.univariate_component(1)
        self.assertTrue(seriesA == TimeSeries.from_times_and_values(self.times1, range(5, 15), columns=["1"]))
        seriesB = self.series1.univariate_component(0).stack(seriesA).stack(self.series1.univariate_component(2))
        self.assertTrue(self.series1 == seriesB)

    def test_add_datetime_attribute(self):
        seriesA = self.series1.add_datetime_attribute('day')
        self.assertEqual(seriesA.width, self.series1.width + 1)
        self.assertTrue(set(seriesA.pd_dataframe().iloc[:, seriesA.width - 1].values.flatten()) == set(range(1, 11)))
        seriesB = self.series3.add_datetime_attribute('day', True)
        self.assertEqual(seriesB.width, self.series3.width + 31)
        self.assertEqual(set(seriesB.pd_dataframe().iloc[:, self.series3.width:].values.flatten()), {0, 1})
        seriesC = self.series1.add_datetime_attribute('month', True)
        self.assertEqual(seriesC.width, self.series1.width + 12)
        seriesD = TimeSeries.from_times_and_values(pd.date_range('20130206', '20130430'), range(84))
        seriesD = seriesD.add_datetime_attribute('month', True)
        self.assertEqual(seriesD.width, 13)
        self.assertEqual(sum(seriesD.values().flatten()), sum(range(84)) + 84)
        self.assertEqual(sum(seriesD.values()[:, 1 + 3]), 30)
        self.assertEqual(sum(seriesD.values()[:, 1 + 1]), 23)

        # test cyclic
        times_month = pd.date_range('20130101', '20140610')
        start = times_month[0]
        end = times_month[-1]

        seriesE = TimeSeries.from_times_and_values(times_month, np.repeat(0.1, len(times_month)))
        seriesF = seriesE.add_datetime_attribute('day', cyclic=True)

        values_sin = seriesF.values()[:, 1]
        values_cos = seriesF.values()[:, 2]

        self.assertTrue(np.allclose(np.add(np.square(values_sin), np.square(values_cos)), 1))
        start_of_month = [pd.Timestamp(year=start.year, month=m, day=1, freq='d') - start for m in range(start.month, end.month)]
        start_of_month_idx = [stamp.days for stamp in start_of_month]

        self.assertTrue(np.allclose(values_sin[start_of_month_idx], 0))
        self.assertTrue(np.allclose(values_cos[start_of_month_idx], 1))

    def test_add_holidays(self):
        times = pd.date_range(start=pd.Timestamp('20201201'), periods=30, freq='D')
        seriesA = TimeSeries.from_times_and_values(times, range(len(times)))

        # testing for christmas and non-holiday in US
        seriesA = seriesA.add_holidays('US')
        last_column = seriesA.pd_dataframe().iloc[:, seriesA.width - 1]
        self.assertEqual(last_column.at[pd.Timestamp('20201225')], 1)
        self.assertEqual(last_column.at[pd.Timestamp('20201210')], 0)
        self.assertEqual(last_column.at[pd.Timestamp('20201226')], 0)

        # testing for christmas and non-holiday in PL
        seriesA = seriesA.add_holidays('PL')
        last_column = seriesA.pd_dataframe().iloc[:, seriesA.width - 1]
        self.assertEqual(last_column.at[pd.Timestamp('20201225')], 1)
        self.assertEqual(last_column.at[pd.Timestamp('20201210')], 0)
        self.assertEqual(last_column.at[pd.Timestamp('20201226')], 1)
        self.assertEqual(seriesA.width, 3)

        # testing hourly time series
        times = pd.date_range(start=pd.Timestamp('20201224'), periods=50, freq='H')
        seriesB = TimeSeries.from_times_and_values(times, range(len(times)))
        seriesB = seriesB.add_holidays('US')
        last_column = seriesB.pd_dataframe().iloc[:, seriesB.width - 1]
        self.assertEqual(last_column.at[pd.Timestamp('2020-12-25 01:00:00')], 1)
        self.assertEqual(last_column.at[pd.Timestamp('2020-12-24 23:00:00')], 0)

    def test_assert_univariate(self):
        with self.assertRaises(AssertionError):
            self.series1._assert_univariate()
        self.series1.univariate_component(0)._assert_univariate()

    def test_first_last_values(self):
        self.assertEqual(self.series1.first_values().tolist(), [0, 5, 10])
        self.assertEqual(self.series3.last_values().tolist(), [10, 20])
        self.assertEqual(self.series1.univariate_component(1).first_values().tolist(), [5])
        self.assertEqual(self.series3.univariate_component(1).last_values().tolist(), [20])




import numpy as np

from .base_test_class import DartsBaseTestClass
from ..utils import timeseries_generation as tg
from ..metrics import mae
from ..logging import get_logger
from darts import TimeSeries
from darts.models import ExponentialSmoothing, ARIMA
from darts.models.forecasting.forecasting_model import GlobalForecastingModel

logger = get_logger(__name__)

try:
    import torch
    from ..models import RNNModel, TCNModel, TransformerModel, BlockRNNModel
    from darts.utils.likelihood_models import (GaussianLikelihood,
                                               PoissonLikelihood,
                                               NegativeBinomialLikelihood,
                                               BernoulliLikelihood,
                                               GammaLikelihood,
                                               GumbelLikelihood,
                                               LaplaceLikelihood,
                                               BetaLikelihood,
                                               ExponentialLikelihood,
                                               DirichletLikelihood,
                                               GeometricLikelihood,
                                               CauchyLikelihood,
                                               ContinuousBernoulliLikelihood,
                                               HalfNormalLikelihood,
                                               LogNormalLikelihood,
                                               WeibullLikelihood)
    TORCH_AVAILABLE = True
except ImportError:
    logger.warning('Torch not available. TCN tests will be skipped.')
    TORCH_AVAILABLE = False

models_cls_kwargs_errs = [
    (ExponentialSmoothing, {}, 0.4),
    (ARIMA, {'p': 1, 'd': 0, 'q': 1}, 0.17)
]

if TORCH_AVAILABLE:
    models_cls_kwargs_errs += [
        (RNNModel, {'input_chunk_length': 2, 'training_length': 10, 'n_epochs': 20, 'random_state': 0,
                    'likelihood': GaussianLikelihood()}, 1.9),
        (TCNModel, {'input_chunk_length': 10, 'output_chunk_length': 5, 'n_epochs': 60, 'random_state': 0,
                    'likelihood': GaussianLikelihood()}, 0.28),
        (BlockRNNModel, {'input_chunk_length': 10, 'output_chunk_length': 5, 'n_epochs': 20, 'random_state': 0,
                         'likelihood': GaussianLikelihood()}, 1),
        (TransformerModel, {'input_chunk_length': 10, 'output_chunk_length': 5, 'n_epochs': 20, 'random_state': 0,
                            'likelihood': GaussianLikelihood()}, 1)
    ]


class ProbabilisticTorchModelsTestCase(DartsBaseTestClass):
    np.random.seed(0)

    constant_ts = tg.constant_timeseries(length=200, value=0.5)
    constant_noisy_ts = constant_ts + tg.gaussian_timeseries(length=200, std=0.1)
    constant_multivar_ts = constant_ts.stack(constant_ts)
    constant_noisy_multivar_ts = constant_noisy_ts.stack(constant_noisy_ts)
    num_samples = 5

    def test_fit_predict_determinism(self):

        for model_cls, model_kwargs, _ in models_cls_kwargs_errs:

            # whether the first predictions of two models initiated with the same random state are the same
            model = model_cls(**model_kwargs)
            model.fit(self.constant_ts)
            pred1 = model.predict(n=10, num_samples=2).values()

            model = model_cls(**model_kwargs)
            model.fit(self.constant_ts)
            pred2 = model.predict(n=10, num_samples=2).values()

            self.assertTrue((pred1 == pred2).all())

            # test whether the next prediction of the same model is different
            pred3 = model.predict(n=10, num_samples=2).values()
            self.assertTrue((pred2 != pred3).any())

    def test_probabilistic_forecast_accuracy(self):
        for model_cls, model_kwargs, err in models_cls_kwargs_errs:
            self.helper_test_probabilistic_forecast_accuracy(model_cls, model_kwargs, err,
                                                             self.constant_ts, self.constant_noisy_ts)
            if issubclass(model_cls, GlobalForecastingModel):
                self.helper_test_probabilistic_forecast_accuracy(model_cls, model_kwargs, err,
                                                                 self.constant_multivar_ts,
                                                                 self.constant_noisy_multivar_ts)

    def helper_test_probabilistic_forecast_accuracy(self, model_cls, model_kwargs, err, ts, noisy_ts):
        model = model_cls(**model_kwargs)
        model.fit(noisy_ts[:100])
        pred = model.predict(n=100, num_samples=100)

        # test accuracy of the median prediction compared to the noiseless ts
        mae_err_median = mae(ts[100:], pred)
        self.assertLess(mae_err_median, err)

        # test accuracy for increasing quantiles between 0.7 and 1 (it should ~decrease, mae should ~increase)
        tested_quantiles = [0.7, 0.8, 0.9, 0.99]
        mae_err = mae_err_median
        for quantile in tested_quantiles:
            new_mae = mae(ts[100:], pred.quantile_timeseries(quantile=quantile))
            self.assertLess(mae_err, new_mae+0.1)
            mae_err = new_mae

        # test accuracy for decreasing quantiles between 0.3 and 0 (it should ~decrease, mae should ~increase)
        tested_quantiles = [0.3, 0.2, 0.1, 0.01]
        mae_err = mae_err_median
        for quantile in tested_quantiles:
            new_mae = mae(ts[100:], pred.quantile_timeseries(quantile=quantile))
            self.assertLess(mae_err, new_mae+0.1)
            mae_err = new_mae

    """ More likelihood tests
    """
    if TORCH_AVAILABLE:
        np.random.seed(42)
        torch.manual_seed(42)

        real_series = TimeSeries.from_values(np.random.randn(100, 2) + [0, 5])
        vals = real_series.all_values()

        real_pos_series = TimeSeries.from_values(np.where(vals > 0, vals, -vals))
        discrete_pos_series = TimeSeries.from_values(np.random.randint(low=0, high=11, size=(100, 2)))
        binary_series = TimeSeries.from_values(np.random.randint(low=0, high=2, size=(100, 2)))
        bounded_series = TimeSeries.from_values(np.random.beta(2, 5, size=(100, 2)))
        simplex_series = bounded_series['0'].stack(1. - bounded_series['0'])

        lkl_series = ((GaussianLikelihood(), real_series, 0.1, 3),
                      (PoissonLikelihood(), discrete_pos_series, 2, 2),
                      (NegativeBinomialLikelihood(), discrete_pos_series, 0.5, 0.5),
                      (BernoulliLikelihood(), binary_series, 0.15, 0.15),
                      (GammaLikelihood(), real_pos_series, 0.3, 0.3),
                      (GumbelLikelihood(), real_series, 0.2, 3),
                      (LaplaceLikelihood(), real_series, 0.3, 4),
                      (BetaLikelihood(), bounded_series, 0.1, 0.1),
                      (ExponentialLikelihood(), real_pos_series, 0.3, 2),
                      (DirichletLikelihood(), simplex_series, 0.3, 0.3),
                      (GeometricLikelihood(), discrete_pos_series, 1, 1),
                      (CauchyLikelihood(), real_series, 3, 10),
                      (ContinuousBernoulliLikelihood(), bounded_series, 0.1, 0.1),
                      (HalfNormalLikelihood(), real_pos_series, 0.3, 8),
                      (LogNormalLikelihood(), real_pos_series, 0.3, 1),
                      (WeibullLikelihood(), real_pos_series, 0.2, 2))

        def test_likelihoods_and_resulting_mean_forecasts(self):
            def _get_avgs(series):
                return np.mean(series.all_values()[:, 0, :]), np.mean(series.all_values()[:, 1, :])

            for lkl, series, diff1, diff2 in self.lkl_series:
                model = RNNModel(input_chunk_length=5, likelihood=lkl)
                model.fit(series, epochs=50)
                pred = model.predict(n=50, num_samples=50)

                avgs_orig, avgs_pred = _get_avgs(series), _get_avgs(pred)
                self.assertLess(abs(avgs_orig[0] - avgs_pred[0]), diff1,
                                'The difference between the mean forecast and the mean series is larger '
                                'than expected on component 0 for distribution {}'.format(lkl))
                self.assertLess(abs(avgs_orig[1] - avgs_pred[1]), diff2,
                                'The difference between the mean forecast and the mean series is larger '
                                'than expected on component 1 for distribution {}'.format(lkl))



import numpy as np

from .base_test_class import DartsBaseTestClass
from darts.models.forecasting.fft import _find_relevant_timestamp_attributes
from ..utils import timeseries_generation as tg


class FFTTestCase(DartsBaseTestClass):

    def helper_relevant_attributes(self, freq, length, period_attributes_tuples):

        # test random walk
        random_walk_ts = tg.random_walk_timeseries(freq=freq, length=length)
        self.assertEqual(_find_relevant_timestamp_attributes(random_walk_ts), set())

        for period, relevant_attributes in period_attributes_tuples:

            # test seasonal period with no noise
            seasonal_ts = tg.sine_timeseries(freq=freq, value_frequency=1 / period, length=length)
            self.assertEqual(_find_relevant_timestamp_attributes(seasonal_ts), relevant_attributes,
                             'failed to recognize season in non-noisy timeseries')

            # test seasonal period with no noise
            seasonal_noisy_ts = seasonal_ts + tg.gaussian_timeseries(freq=freq, length=length)
            self.assertEqual(_find_relevant_timestamp_attributes(seasonal_noisy_ts), relevant_attributes,
                             'failed to recognize season in noisy timeseries')

    def test_find_relevant_timestamp_attributes(self):

        np.random.seed(0)

        # monthly frequency
        self.helper_relevant_attributes('M', 150, [
            (12, {'month'}),
        ])

        # daily frequency
        self.helper_relevant_attributes('D', 1000, [
            (365, {'month', 'day'}),
            (30, {'day'}),
            (7, {'weekday'})
        ])

        # hourly frequency
        self.helper_relevant_attributes('H', 3000, [
            (730, {'day', 'hour'}),
            (168, {'weekday', 'hour'}),
            (24, {'hour'})
        ])

        # minutely frequency
        self.helper_relevant_attributes('min', 5000, [
            (1440, {'hour', 'minute'}),
            (60, {'minute'}),
        ])



import numpy as np
import pandas as pd

from .base_test_class import DartsBaseTestClass
from .. import TimeSeries
from ..utils.statistics import check_seasonality


class TimeSeriesTestCase(DartsBaseTestClass):

    def test_check_seasonality(self):
        pd_series = pd.Series(range(50), index=pd.date_range('20130101', '20130219'))
        pd_series = pd_series.map(lambda x: np.sin(x * np.pi / 3 + np.pi / 2))
        series = TimeSeries.from_series(pd_series)

        self.assertEqual((True, 6), check_seasonality(series))
        self.assertEqual((False, 3), check_seasonality(series, m=3))

        with self.assertRaises(AssertionError):
            check_seasonality(series.stack(series))



from darts.timeseries import TimeSeries
from darts.metrics import mape

import pandas as pd
import numpy as np
import xarray as xr
import unittest
from matplotlib import pyplot as plt

from darts.tests.base_test_class import DartsBaseTestClass
from darts.utils import timeseries_generation as tg
from darts.utils import missing_values as mv
from darts.dataprocessing import dtw
from darts.metrics import mae, mape, dtw_metric
from darts.datasets import EnergyDataset


def _series_from_values(values):
    return TimeSeries.from_values(np.array(values) if isinstance(values, list) else values)


class DynamicTimeWarpingTestCase(DartsBaseTestClass):
    length = 20
    freq = 1 / length
    series1 = tg.sine_timeseries(length=length, value_frequency=freq, value_phase=0, value_y_offset=5)
    series2 = tg.sine_timeseries(length=length, value_frequency=freq, value_phase=np.pi / 4, value_y_offset=5)

    def test_shift(self):
        input1 = [1, 1, 1, 1, 1.2, 1.4, 1.2, 1, 1, 1, 1, 1, 1, 1.2, 1.4, 1.6, 1.8, 1.6, 1.4, 1.2, 1, 1]
        input2 = [1] + input1[:-1]

        expected_path = [(0, 0)] + list((i - 1, i) for i in range(1, len(input1))) + [
            (len(input1) - 1, len(input2) - 1)]

        series1 = _series_from_values(input1)
        series2 = _series_from_values(input2)

        exact_alignment = dtw.dtw(series1, series2, multi_grid_radius=-1)

        self.assertEqual(exact_alignment.distance(), 0, "Minimum cost between two shifted series should be 0")
        self.assertTrue(np.array_equal(exact_alignment.path(), expected_path), "Incorrect path")

    def test_multi_grid(self):
        size = 2 ** 5 - 1  # test odd size
        freq = 1 / size
        input1 = np.cos(np.arange(size) * 2 * np.pi * freq)
        input2 = np.sin(np.arange(size) * 2 * np.pi * freq) + 0.1 * np.random.random(size=size)

        series1 = _series_from_values(input1)
        series2 = _series_from_values(input2)

        exact_distance = dtw.dtw(series1, series2, multi_grid_radius=-1).distance()
        approx_distance = dtw.dtw(series1, series2, multi_grid_radius=1).distance()

        self.assertAlmostEqual(exact_distance, approx_distance, 3)

    def test_sakoe_chiba_window(self):
        window = 2
        alignment = dtw.dtw(self.series1, self.series2, window=dtw.SakoeChiba(window_size=2))
        path = alignment.path()

        for i, j in path:
            self.assertGreaterEqual(window, abs(i - j))

    def test_itakura_window(self):
        n = 6
        m = 5
        slope = 1.5

        window = dtw.Itakura(max_slope=slope)
        window.init_size(n, m)

        cells = list(window)
        self.assertEqual(cells, [(1, 1), (1, 2), (2, 1), (2, 2), (2, 3), (3, 1), (3, 2), (3, 3), (3, 4), (4, 2), (4, 3),
                                 (4, 4), (5, 2), (5, 3), (5, 4), (5, 5), (6, 4), (6, 5)])

        sizes = [(10, 43), (543, 45), (34, 11)]

        for n, m in sizes:
            slope = m / n + 1

            series1 = tg.sine_timeseries(length=n, value_frequency=1 / n, value_phase=0)
            series2 = tg.sine_timeseries(length=m, value_frequency=1 / m, value_phase=np.pi / 4)

            dist = dtw.dtw(series1, series2, window=dtw.Itakura(slope)).mean_distance()
            self.assertGreater(1, dist)

    def test_warp(self):
        # Support different time dimension names
        xa1 = self.series1.data_array().rename({"time": "time1"})
        xa2 = self.series2.data_array().rename({"time": "time2"})

        series1 = TimeSeries.from_xarray(xa1)
        series2 = TimeSeries.from_xarray(xa2)

        alignment = dtw.dtw(series1, series2)

        warped1, warped2 = alignment.warped()
        self.assertAlmostEqual(alignment.mean_distance(), mae(warped1, warped2))

        """
        See DTWAlignment.warped for why this functionality is currently disabled

        #Mutually Exclusive Option
        with self.assertRaises(ValueError):
            alignment.warped(take_dates=True, range_index=True)

        #Take_dates does not support indexing by RangeIndex
        with self.assertRaises(ValueError):
            xa3 = xa1.copy()
            xa3["time1"] = pd.RangeIndex(0, len(self.series1))

            dtw.dtw(TimeSeries.from_xarray(xa3), series2).warped(take_dates=True)


        warped1, warped2 = alignment.warped(take_dates=True)
        self.assertTrue(np.all(warped1.time_index == warped2.time_index))
        """

    def test_metric(self):
        metric1 = dtw_metric(self.series1, self.series2, metric=mae)
        metric2 = dtw_metric(self.series1, self.series2, metric=mape)

        self.assertGreater(0.5, metric1)
        self.assertGreater(5, metric2)

    def test_nans(self):
        with self.assertRaises(ValueError):
            series1 = _series_from_values([np.nan, 0, 1, 2, 3])
            series2 = _series_from_values([0, 1, 2, 3, 4])

            dtw.dtw(series1, series2)

    def test_plot(self):
        align = dtw.dtw(self.series2, self.series1)
        align.plot()
        align.plot_alignment()

    def test_multivariate(self):
        n = 2

        values1 = np.repeat(self.series1.univariate_values(), n)
        values2 = np.repeat(self.series2.univariate_values(), n)

        values1 = values1.reshape((-1, n))
        values2 = values2.reshape((-1, n))

        multi_series1 = TimeSeries.from_values(values1)
        multi_series2 = TimeSeries.from_values(values2)

        radius = 2

        alignment_uni = dtw.dtw(self.series1, self.series2, multi_grid_radius=radius)
        alignment_multi = dtw.dtw(multi_series1, multi_series2, multi_grid_radius=radius)

        self.assertTrue(np.all(alignment_uni.path() == alignment_multi.path()))


# MINI_BENCHMARK
def _dtw_exact():
    dtw.dtw(series1, series2, multi_grid_radius=-1).distance()


def _dtw_multigrid():
    dtw.dtw(series1, series2, multi_grid_radius=1).distance()


def _benchmark_dtw():
    size1 = 2 ** 10
    size2 = 2 ** 10
    freq1 = 1 / size1
    freq2 = 1 / size2
    input1 = np.cos(np.arange(size1) * 2 * np.pi * freq1)
    input2 = np.sin(np.arange(size2) * 2 * np.pi * freq2) + 0.1 * np.random.random(size=size2)

    global series1, series2

    series1 = _series_from_values(input1)
    series2 = _series_from_values(input2)

    import cProfile
    cProfile.run("_dtw_exact()", sort="tottime")
    cProfile.run("_dtw_multigrid()", sort="tottime")


if __name__ == '__main__':
    unittest.main()
